#include <iostream>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <time.h>
#include <mm.h>
#include <main.h>

using namespace std;

const int BUFFER_SIZE = 1000;
const char TICKER_SIZE = 4;

typedef struct {
  char   ticker[TICKER_SIZE];
  double price;
} item;


char* getTickerSymbol();
// Precondition: none
// Postcondition: returns a TICKER_SIZE character C-string the first 3 characters of which are random 
//                letters 'A'-'Z'.  The last character is '\0'.
// NOTE:          You must call free() to release the C-string returned when you are done with
//                it or you'll end up with a memory leak.

// This program demonstrates a solution to the producer consumer problem using a pipe as the
// as the communication mechanism.  The description of the problem was taken from 
// "Operating System Concepts: 6th Edition By Silbershatz, Galvin, Gagne".

int main(void)
{
  int thePipe[2];

  // Create a single pipe.  The pipe has two ends:
  // - thePipe[0] receiving end - messages are taken out of this end
  // - thePipe[1] sending   end - messages are placed in this end
  // NOTE: This MUST BE DONE BEFORE YOU FORK!
  pipe(thePipe);
  
  // create a child
  int pid = fork();

  // Error forking
  if (pid < 0) {
    cerr << "main: Fork failed!" << endl;
    exit(-1);
  }

  /////////////////////////////////////////////////////////////////////
  // Child process is the CONSUMER
  /////////////////////////////////////////////////////////////////////
  if (pid ==0) {
    
    // Pipe manipulation for the RECEIVER
    // close thePipe[1] we're not using the sending end
    close(thePipe[1]);

    while (true) {
   
      // Read item from pipe
      item nextConsumed;
      read(thePipe[0],&nextConsumed,sizeof(item));

      // Output item
      cout << "child: " << nextConsumed.ticker << " " << nextConsumed.price << endl;

    }
  }

  /////////////////////////////////////////////////////////////////////
  // Parent process is the PRODUCER
  /////////////////////////////////////////////////////////////////////
  else {
    int count=0;

    // Pipe manipulation for the PRODUCER
    // close thePipe[0] we're not using the receiving end
    close(thePipe[0]);

    while (true) {

      // Create item
      item nextProduced;
      char* deleteme;
      strcpy(nextProduced.ticker,deleteme=getTickerSymbol());
      nextProduced.price = count;
      delete [] deleteme;
      count++;

      // Send item
      write(thePipe[1],&nextProduced,sizeof(item));
    }
  }

  return 1;
}

char* getTickerSymbol()
{
  static bool firstTime = true;

  if (firstTime) {
    firstTime = false;
    srand(time(0));
  }

  char* result = new char[TICKER_SIZE];
  
  result[0]='A'+rand() % 26;
  result[1]='A'+rand() % 26;
  result[2]='A'+rand() % 26;
  result[3]='\0';

  return result;
}
